from typing import Any, List, Dict, Tuple, Optional

from deprecated.classic import deprecated
from drb.core import DrbNode
from drb.nodes.abstract_node import AbstractNode
from drb.core.factory import DrbFactory
from drb.core.path import ParsedPath
import zarr

from drb.exceptions.zarr import DrbZarrNodeException
from .zarr_data_set_node import DrbZarrDataSetNode


class DrbZarrNode(AbstractNode):
    """
    This node is used to instantiate a DrbZarrNode
    from another implementation of drb such as file.

    Parameters:
        base_node (DrbNode): the base node of this node.
    """

    def __init__(self, base_node: DrbNode):
        super().__init__()

        self.base_node = base_node
        self._children: List[DrbNode] = None

    @property
    def parent(self) -> Optional[DrbNode]:
        return self.base_node.parent

    @property
    def path(self) -> ParsedPath:
        return self.base_node.path

    @property
    def name(self) -> str:
        return self.base_node.name

    @property
    def namespace_uri(self) -> Optional[str]:
        return self.base_node.namespace_uri

    @property
    def value(self) -> Optional[Any]:
        return self.base_node.value

    @property
    @deprecated(version="1.2.0", reason="drb core deprecation since 2.1.0")
    def attributes(self) -> Dict[Tuple[str, str], Any]:
        return self.base_node.attributes

    def impl_capabilities(self) -> List[type]:
        return self.base_node.impl_capabilities()

    @deprecated(version="1.2.0", reason="drb core deprecation since 2.1.0")
    def get_attribute(self, name: str, namespace_uri: str = None) -> Any:
        return self.base_node.get_attribute(name, namespace_uri)

    @property
    @deprecated(version="1.2.0", reason="drb core deprecation since 2.1.0")
    def children(self) -> List[DrbNode]:
        if self._children is None:
            try:
                root_data_set = zarr.open(self.base_node.path.name)
                root_node = DrbZarrDataSetNode.create_node_from_data_set(
                    self, root_data_set
                )
                self._children = [root_node]
            except Exception as e:
                raise DrbZarrNodeException(
                    f"Unable to read zarr file" f" {self.name} "
                ) from e
        return self._children

    def has_child(self, name: str = None, namespace: str = None) -> bool:
        if namespace is not None:
            return False
        return super().has_child(name, namespace)

    def has_impl(self, impl: type) -> bool:
        return self.base_node.has_impl(impl)

    def get_impl(self, impl: type, **kwargs) -> Any:
        return self.base_node.get_impl(impl)

    def close(self):
        self.base_node.close()

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __delitem__(self, key):
        raise NotImplementedError


class DrbZarrFactory(DrbFactory):
    def _create(self, node: DrbNode) -> DrbNode:
        if isinstance(node, DrbZarrNode):
            return node
        return DrbZarrNode(base_node=node)



import os
import sys
import unittest
import uuid
from pathlib import Path

from drb.core.factory import FactoryLoader
from drb.nodes.logical_node import DrbLogicalNode
from drb.topics.dao import ManagerDao
from drb.topics.topic import TopicCategory

from drb.drivers.zarr import DrbZarrFactory


class TestDrbZarrSignature(unittest.TestCase):
    mock_pkg = None
    fc_loader = None
    topic_loader = None
    zarr_id = uuid.UUID("56e6509c-3666-11ec-8d3d-0242ac130003")
    current_path = Path(os.path.dirname(os.path.realpath(__file__)))

    zarr_ok2 = current_path / "files" / "sample.zarr"
    not_zarr_files = current_path / "files" / "empty.file"

    @classmethod
    def setUpClass(cls) -> None:
        cls.fc_loader = FactoryLoader()
        cls.topic_loader = ManagerDao()

    def test_impl_loading(self):
        factory_name = "zarr"

        factory = self.fc_loader.get_factory(factory_name)
        self.assertIsNotNone(factory)
        self.assertIsInstance(factory, DrbZarrFactory)

        topic = self.topic_loader.get_drb_topic(self.zarr_id)
        self.assertIsNotNone(factory)
        self.assertEqual(self.zarr_id, topic.id)
        self.assertEqual("zarr", topic.label)
        self.assertIsNone(topic.description)
        self.assertEqual(TopicCategory.FORMATTING, topic.category)
        self.assertEqual(factory_name, topic.factory)

    def test_impl_signatures(self):
        topic = self.topic_loader.get_drb_topic(self.zarr_id)

        node = DrbLogicalNode(self.zarr_ok2)
        self.assertTrue(topic.matches(node))

        node = DrbLogicalNode(self.not_zarr_files)
        self.assertFalse(topic.matches(node))



import enum
from abc import ABC
from typing import Any, List, Optional, Union, Dict, Tuple

import drb.topics.resolver as resolver
import zarr
from deprecated.classic import deprecated

from drb.core import DrbNode
from drb.nodes.abstract_node import AbstractNode
from drb.exceptions.core import DrbNotImplementationException, DrbException
from drb.core.path import ParsedPath


class DrbZarrAttributeNames(enum.Enum):
    """
    A Boolean that tell if the zarr is read only.
    """

    READ_ONLY = "read_only"


class DrbZarrDataSetNode(AbstractNode, ABC):
    """
    This node will be inherited by DrbZarrGroupNode and DrbZarrArrayNode.

    Parameters:
        parent (DrbNode): The parent of the node.
        name (str): The name of the node.
        data_set (Union[zarr.hierarchy.Group, zarr.core.Array]):
                 the dataset of the node
                 corresponding to a zarr group
                 otherwise a zarr array.
    """

    def __init__(
        self,
        parent: DrbNode,
        name: str,
        data_set: Union[zarr.hierarchy.Group, zarr.core.Array],
    ):
        super().__init__()

        self._data_set = data_set
        self.parent: DrbNode = parent
        self.name = name
        self._attributes: Dict[Tuple[str, str], Any] = None
        self._available_impl = [
            data_set.__class__
        ]
        self.__init_attributes()

    def __init_attributes(self):
        for key in self._data_set.attrs.keys():
            self @= (key, self._data_set.attrs[key])
        self @= (DrbZarrAttributeNames.READ_ONLY.value,
                 self._data_set.read_only)

    def get_impl(self, impl: type, **kwargs) -> Any:
        if self.has_impl(impl):
            return self._data_set
        raise DrbNotImplementationException(
            f"no {impl} " f"implementation found"
        )

    @staticmethod
    def create_node_from_data_set(parent, data_set):
        name = data_set.name
        if name and name[0] == "/":
            name = name[1:]
        if not name:
            name = "."
        if isinstance(data_set, zarr.hierarchy.Group):
            node = DrbZarrGroupNode(parent, name, data_set)
        else:
            node = DrbZarrArrayNode(parent, name, data_set)
        return node

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __delitem__(self, key):
        raise NotImplementedError


class DrbZarrGroupNode(DrbZarrDataSetNode):
    """
    This node is used to organize the data in a zarr architecture,
    he contains a group of zarr array in his children
    or another DrbZarrGroupNode.

    Parameters:
        parent (DrbNode): The parent of the node.
        name (str): The name of the node.
        data_set (Union[zarr.hierarchy.Group, zarr.core.Array]):
                 the dataset of the node
                 corresponding to a zarr group
                 otherwise a zarr array.
    """

    def __init__(
        self,
        parent: DrbNode,
        name: str,
        data_set: Union[zarr.hierarchy.Group, zarr.core.Array],
    ):
        super().__init__(parent, name, data_set)
        self._children: List[DrbNode] = None

    @property
    @deprecated(version="1.2.0", reason="drb core deprecation since 2.1.0")
    @resolver.resolve_children
    def children(self) -> List[DrbNode]:
        if self._children is None:
            self._children = []
            self._data_set.visitvalues(self.add_data_set_children)
        return self._children

    def add_data_set_children(self, data_set):
        """
        Add the dataset given in argument to the List of children.

        Parameters:
            data_set : a data set to be added in the children List.
        """
        child = DrbZarrDataSetNode.create_node_from_data_set(self, data_set)
        self._children.append(child)

    def _get_named_child(
        self,
        name: str,
        namespace_uri: str = None,
        occurrence: Union[int, slice] = 0,
    ) -> Union[DrbNode, List[DrbNode]]:
        if self._children is None:
            if namespace_uri is None:
                data_set = self._data_set[name]
                return [self.create_node_from_data_set(self, data_set)][
                    occurrence
                ]

            raise DrbException(
                f"No child found having name: {name} and"
                f" namespace: {namespace_uri}"
            )
        else:
            return super()._get_named_child(name, namespace_uri, occurrence)


class DrbZarrArrayNode(DrbZarrDataSetNode):
    """
    This node is used to represent one or a set of values
    contained in a Zarr array.

    Parameters:
        parent (DrbNode): The parent of the node.
        name (str): The name of the node.
        data_set (Union[zarr.hierarchy.Group, zarr.core.Array]):
                 the dataset of the node
                 corresponding to a zarr group
                 otherwise a zarr array.
    """

    def __init__(
        self,
        parent: DrbNode,
        name: str,
        data_set: Union[zarr.hierarchy.Group, zarr.core.Array],
    ):
        super().__init__(parent, name, data_set)

    @property
    @deprecated(version="1.2.0", reason="drb core deprecation since 2.1.0")
    def children(self) -> List[DrbNode]:
        return []



from .zarr_data_set_node import DrbZarrArrayNode, DrbZarrGroupNode
from .zarr_node_factory import DrbZarrFactory, DrbZarrNode
from . import _version

__version__ = _version.get_versions()["version"]

del _version

__all__ = [
    "DrbZarrFactory",
    "DrbZarrNode",
    "DrbZarrNode",
    "DrbZarrArrayNode",
    "DrbZarrGroupNode",
]
